#ifndef TYPON_CORE_FORK_HPP_INCLUDED
#define TYPON_CORE_FORK_HPP_INCLUDED

#include <coroutine>
#include <cstdint>

#include <typon/core/forked.hpp>
#include <typon/core/parent.hpp>
#include <typon/core/result.hpp>
#include <typon/core/scheduler.hpp>


namespace typon
{

  template <typename T = void>
  struct [[nodiscard]] Fork
  {
    struct promise_type;
    using u64 = Parent::u64;

    std::coroutine_handle<promise_type> _coroutine;

    Fork(std::coroutine_handle<promise_type> coroutine) noexcept
      : _coroutine(coroutine)
    {}

    Fork(const Fork &) = delete;
    Fork& operator=(const Fork &) = delete;

    Fork(Fork &&) noexcept = default;
    Fork& operator=(Fork &&) noexcept = default;

    struct promise_type : Result<T>
    {
      Parent * _parent;
      ForkList::Node _node;

      Fork get_return_object() noexcept
      {
        return { std::coroutine_handle<promise_type>::from_promise(*this) };
      }

      std::suspend_always initial_suspend() noexcept
      {
        return {};
      }

      auto final_suspend() noexcept
      {
        struct awaitable : std::suspend_always
        {
          std::coroutine_handle<> await_suspend(std::coroutine_handle<promise_type> coroutine) noexcept
          {
            auto parent = coroutine.promise()._parent;
            if (Scheduler::pop())
            {
              return parent->_coroutine;
            }
            u64 n = parent->_n.fetch_sub(1, std::memory_order_acq_rel);
            if (n == 1)
            {
              return parent->continuation();
            }
            return std::noop_coroutine();
          }
        };

        return awaitable {};
      }
    };

    struct awaitable
    {
      std::coroutine_handle<promise_type> _coroutine;
      u64 _thefts;

      awaitable(std::coroutine_handle<promise_type> coroutine) noexcept
        : _coroutine(coroutine)
      {}

      bool await_ready() noexcept
      {
        return false;
      }

      template <typename Promise>
      auto await_suspend(std::coroutine_handle<Promise> continuation) noexcept
      {
        Parent * parent = &(continuation.promise()._data);
        _coroutine.promise()._parent = parent;
        _thefts = parent->_thefts;

        std::coroutine_handle<> on_stack_handle = _coroutine;
        Scheduler::push(parent);
        return on_stack_handle;
      }

      auto await_resume()
      {
        auto parent = _coroutine.promise()._parent;
        bool stolen = parent->_thefts > _thefts;
        if (stolen)
        {
          parent->_children.insert(_coroutine);
        }
        return Forked<T>(_coroutine, !stolen);
      }
    };

    auto operator co_await() &&
    {
      return awaitable { _coroutine };
    }
  };

}

#endif // TYPON_FORK_HPP_INCLUDED
