#ifndef TYPON_CORE_FORKED_HPP_INCLUDED
#define TYPON_CORE_FORKED_HPP_INCLUDED

#include <coroutine>
#include <memory>
#include <type_traits>
#include <utility>

#include <typon/fundamental/scope.hpp>

#include <typon/core/result.hpp>


namespace typon
{

  template <typename T>
  struct Forked
  {
    using value_type = T;

    Result<T> * _result = nullptr;
    union
    {
      T _value;
    };

    template <typename Promise>
    Forked(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        std::construct_at(std::addressof(_value), coroutine.promise().get());
      }
      else
      {
        _result = &(coroutine.promise());
      }
    }

    Forked(Forked && other) noexcept(std::is_nothrow_move_constructible_v<T>)
    {
      _result = other._result;
      if (!_result)
      {
        std::construct_at(std::addressof(_value), std::move(other._value));
      }
    }

    Forked& operator=(Forked && other)
      noexcept(std::is_nothrow_move_constructible_v<T>)
    {
      if (this != &other)
      {
        Forked old { std::move(*this) };
        _result = other._result;
        if (!_result)
        {
          std::construct_at(std::addressof(_value), std::move(other._value));
        }
      }
      return *this;
    }

    ~Forked()
    {
      if (!_result)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    T get()
    {
      if (_result)
      {
        return _result->get();
      }
      return _value;
    }
  };


  template <typename T>
    requires requires { sizeof(T); } && (sizeof(T) > 2 * sizeof(void*))
  struct Forked<T>
  {
    using value_type = T;

    Result<T> * _result = nullptr;
    std::coroutine_handle<> _coroutine;

    template <typename Promise>
    Forked(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      _result = &(coroutine.promise());
      if (ready)
      {
        _coroutine = coroutine;
        if (coroutine.promise()._exception)
        {
          std::rethrow_exception(coroutine.promise()._exception);
        }
      }
    }

    Forked(const Forked &) = delete;
    Forked& operator=(const Forked &) = delete;

    Forked(Forked && other) noexcept
      : _result(other._result)
      , _coroutine(std::exchange(other._coroutine, nullptr))
    {}

    Forked& operator=(Forked && other) noexcept
    {
      std::swap(_coroutine, other._coroutine);
      std::swap(_result, other._result);
      return *this;
    }

    ~Forked()
    {
      if (_coroutine)
      {
        _coroutine.destroy();
      }
    }

    T get()
    {
      return _result->get();
    }
  };


  template <typename T>
    requires std::is_trivially_copyable_v<T>
  struct Forked<T>
  {
    using value_type = T;

    Result<T> * _result = nullptr;
    union
    {
      value_type _value;
    };

    template <typename Promise>
    Forked(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        std::construct_at(std::addressof(_value), coroutine.promise().get());
      }
      else
      {
        _result = &(coroutine.promise());
      }
    }

    Forked(Forked && other) = default;
    Forked& operator=(Forked && other) = default;

    ~Forked()
    {
      if (!_result)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    T get()
    {
      if (_result)
      {
        return _result->get();
      }
      return _value;
    }
  };


  template <typename T>
  struct Forked<T&>
  {
    using value_type = T&;

    Result<T> * _result = nullptr;
    T * _value;

    template <typename Promise>
    Forked(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        _value = std::addressof(coroutine.promise().get());
      }
      else
      {
        _result = &(coroutine.promise());
      }
    }

    T& get()
    {
      if (_result)
      {
        return _result->get();
      }
      return *_value;
    }
  };


  template <>
  struct Forked<void>
  {
    using value_type = void;

    Result<void> * _result = nullptr;

    template <typename Promise>
    Forked(std::coroutine_handle<Promise> coroutine, bool ready)
    {
      if (ready)
      {
        fdt::defer defer { [&coroutine]() { coroutine.destroy(); } };
        coroutine.promise().get();
      }
      else
      {
        _result = &(coroutine.promise());
      }
    }

    void get()
    {
      if (_result)
      {
        _result->get();
      }
    }
  };

}


#endif // TYPON_FORKED_HPP_INCLUDED
