#ifndef TYPON_CORE_RESULT_HPP_INCLUDED
#define TYPON_CORE_RESULT_HPP_INCLUDED


#include <exception>
#include <memory>
#include <type_traits>
#include <utility>


namespace typon
{

  template <typename T>
  struct Result
  {
    using value_type = T;

    bool _valid = false;
    std::exception_ptr _exception;
    union
    {
      T _value;
    };

    ~Result()
    {
      if (_valid)
      {
        std::destroy_at(std::addressof(_value));
      }
    }

    template <typename U>
    void return_value(U&& expr) noexcept(std::is_nothrow_constructible_v<T, U&&>)
    {
      std::construct_at(std::addressof(_value), std::forward<U>(expr));
      _valid = true;
    }

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    T& get() &
    {
      if (_exception)
      {
        std::rethrow_exception(std::exchange(_exception, nullptr));
      }
      return _value;
    }

    T&& get() &&
    {
      if (_exception)
      {
        std::rethrow_exception(std::exchange(_exception, nullptr));
      }
      return std::move(_value);
    }

  };


  template <typename T>
  struct Result<T&>
  {
    using value_type = T&;

    T* _value;
    std::exception_ptr _exception;

    void return_value(T& expr) noexcept
    {
      _value = std::addressof(expr);
    }

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    T& get() &
    {
      if (_exception)
      {
        std::rethrow_exception(std::exchange(_exception, nullptr));
      }
      return *_value;
    }
  };


  template <>
  struct Result<void>
  {
    using value_type = void;

    std::exception_ptr _exception;

    void return_void() noexcept {}

    void unhandled_exception() noexcept
    {
      _exception = std::current_exception();
    }

    void get()
    {
      if (_exception)
      {
        std::rethrow_exception(std::exchange(_exception, nullptr));
      }
    }
  };

}

#endif // TYPON_RESULT_HPP_INCLUDED
