#ifndef TYPON_CORE_SCHEDULER_HPP_INCLUDED
#define TYPON_CORE_SCHEDULER_HPP_INCLUDED

#include <atomic>
#include <bit>
#include <coroutine>
#include <cstdint>
#include <exception>
#include <thread>
#include <utility>
#include <vector>

#include <typon/fundamental/event_count.hpp>
#include <typon/fundamental/garbage_collector.hpp>
#include <typon/fundamental/random.hpp>

#include <typon/core/continuation.hpp>
#include <typon/core/pool.hpp>
#include <typon/core/stack.hpp>


namespace typon
{

  struct Scheduler
  {
    using uint = unsigned int;
    using u64 = std::uint_fast64_t;

    std::atomic_bool _done {false};
    const uint _concurrency;
    const uint _mask;
    std::atomic<uint> _thieves = 0;
    std::atomic<u64> _potential = 0;
    fdt::lock_free::event_count<> _notifyer;
    std::vector<Pool> _pool;
    std::vector<Stack *> _stack;
    std::vector<std::thread> _thread;
    fdt::lock_free::garbage_collector _gc;

    static inline thread_local uint thread_id;

    static Scheduler & get() noexcept
    {
      static Scheduler scheduler {std::thread::hardware_concurrency()};
      return scheduler;
    }

    static void schedule(std::coroutine_handle<> coroutine) noexcept
    {
      Pool & pool = get().random();
      Stack * stack = new Stack(Stack::READY);
      {
        auto lock_guard = pool.lock_guard();
        pool.add(stack);
      }
      stack->_coroutine = coroutine;
      get()._potential.fetch_add(1);
      get()._notifyer.notify_one();
    }

    static void push(Continuation * task) noexcept
    {
      get()._stack[thread_id]->push(task);
    }

    static bool pop() noexcept
    {
      bool result = get()._stack[thread_id]->pop();
      if (auto garbage = get()._stack[thread_id]->reclaim())
      {
        get()._gc.retire(garbage);
      }
      return result;
    }

    static Continuation * peek() noexcept
    {
      return get()._stack[thread_id]->peek();
    }

    static auto suspend(std::coroutine_handle<> coroutine) noexcept
    {
      Stack * stack = std::exchange(get()._stack[thread_id], nullptr);
      stack->_coroutine = coroutine;
      stack->_state.store(Stack::WAITING);
      return stack;
    }

    static void enable(Stack * stack) noexcept
    {
      auto state = stack->_state.exchange(Stack::READY);
      if (state == Stack::EMPTY)
      {
        Pool & pool = get().random();
        {
          auto lock_guard = pool.lock_guard();
          pool.add(stack);
        }
        get()._potential.fetch_add(1);
        get()._notifyer.notify_one();
      }
    }

    Scheduler(uint concurrency) noexcept
      : _concurrency(concurrency)
      , _mask((1 << std::bit_width(concurrency)) - 1)
      , _pool(this->_mask + 1)
      , _stack(concurrency, nullptr)
      , _gc(concurrency)
    {
      for (uint id = 0; id < concurrency; id++)
      {
        _thread.emplace_back([this, id]() {
          thread_id = id;
          for(;;)
          {
            std::coroutine_handle<> coroutine;
            if (!wait_for_work(coroutine))
            {
              break;
            }
            coroutine.resume();
          }
        });
      }
    }

    bool wait_for_work(std::coroutine_handle<> & coroutine) noexcept
    {
      for(;;)
      {
        _thieves.fetch_add(1);
        find_work(coroutine);
        if (coroutine)
        {
          prepare_stack();
          if (_thieves.fetch_sub(1) == 1)
          {
            _notifyer.notify_one();
          }
          return true;
        }
        auto key = _notifyer.prepare_wait();
        if (_done.load())
        {
          _notifyer.cancel_wait();
          _notifyer.notify_all();
          _thieves.fetch_sub(1);
          return false;
        }
        if (_thieves.fetch_sub(1) == 1)
        {
          if (_potential.load() > 0)
          {
            _notifyer.cancel_wait();
            continue;
          }
        }
        _notifyer.wait(key);
      }
    }

    void find_work(std::coroutine_handle<> & coroutine) noexcept
    {
      auto epoch = _gc.epoch(thread_id);
      for (uint i = 0; i < _concurrency * 2 + 1; i++)
      {
        Pool & pool = random();
        if (!pool.try_lock())
        {
          continue;
        }
        auto lock_guard = pool.adopt_lock_guard();
        auto size = pool.size();
        if (!size)
        {
          continue;
        }
        auto index = size > 1 ? fdt::random::random() % size : 0;
        auto stack = pool.get(index);
        auto state = stack->_state.load();
        if (state == Stack::ACTIVE)
        {
          if (auto task = stack->steal())
          {
            task->_thefts++;
            coroutine = task->_coroutine;
          }
          return;
        }
        if (state == Stack::WAITING)
        {
          if (auto task = stack->pop_top())
          {
            task->_thefts++;
            coroutine = task->_coroutine;
            return;
          }
          if (stack->_state.compare_exchange_strong(state, Stack::EMPTY))
          {
            _potential.fetch_sub(1);
            pool.remove(index);
            return;
          }
          adopt_stack(stack);
          coroutine = stack->_coroutine;
          return;
        }
        if (state == Stack::READY)
        {
          adopt_stack(stack);
          coroutine = stack->_coroutine;
          return;
        }
        if (state == Stack::DONE)
        {
          pool.remove(index);
          delete stack;
          return;
        }
      }
    }

    void adopt_stack(Stack * stack) noexcept
    {
      stack->_state.store(Stack::ACTIVE);
      if (auto old = std::exchange(_stack[thread_id], stack))
      {
        _potential.fetch_sub(1);
        old->_state.store(Stack::DONE);
      }
    }

    void prepare_stack() noexcept
    {
      if (!_stack[thread_id])
      {
        auto stack = _stack[thread_id] = new Stack(Stack::ACTIVE);
        Pool & pool = random();
        {
          auto lock_guard = pool.lock_guard();
          pool.add(stack);
        }
        _potential.fetch_add(1);
        _notifyer.notify_one();
      }
    }

    Pool & random() noexcept
    {
      return _pool[fdt::random::random() & _mask];
    }

    ~Scheduler() noexcept
    {
      _done.store(true);
      _notifyer.notify_all();
      for (auto & t : _thread)
      {
        t.join();
      }
    }
  };

}

#endif // TYPON_SCHEDULER_HPP_INCLUDED
