#ifndef TYPON_FUNDAMENTAL_STACK_HPP_INCLUDED
#define TYPON_FUNDAMENTAL_STACK_HPP_INCLUDED

#include <atomic>
#include <cstdint>
#include <memory>
#include <type_traits>
#include <utility>

#include <typon/fundamental/ring_buffer.hpp>

#include <typon/core/continuation.hpp>


namespace typon
{

  struct Stack
  {
    using ring_buffer = fdt::lock_free::ring_buffer<Continuation *>;
    using u64 = ring_buffer::u64;

    using enum std::memory_order;

    enum State : unsigned char { ACTIVE, WAITING, EMPTY, READY, DONE };

    std::atomic<u64> _top {1};
    std::atomic<u64> _bottom {1};
    std::atomic<ring_buffer *> _buffer;
    std::atomic<State> _state;
    std::coroutine_handle<> _coroutine;

    Stack(State state) noexcept
      : _buffer(new ring_buffer(3))
      , _state(state)
    {}

    ~Stack()
    {
      delete _buffer.load(relaxed);
    }

    void push(Continuation * x) noexcept
    {
      u64 bottom = _bottom.load(relaxed);
      u64 top = _top.load(acquire);
      ring_buffer * buffer = _buffer.load(relaxed);
      if (bottom - top > buffer->capacity() - 1)
      {
        buffer = buffer->grow(top, bottom);
        _buffer.store(buffer);
      }
      buffer->put(bottom, x);
      std::atomic_thread_fence(release);
      _bottom.store(bottom + 1, relaxed);
    }

    bool pop() noexcept
    {
      u64 bottom = _bottom.load(relaxed) - 1;
      _bottom.store(bottom, relaxed);
      std::atomic_thread_fence(seq_cst);
      u64 top = _top.load(relaxed);
      if (top < bottom)
      {
        return true;
      }
      if (top == bottom)
      {
        bool win = _top.compare_exchange_strong(top, top + 1, seq_cst, relaxed);
        _bottom.store(bottom + 1, relaxed);
        return win;
      }
      _bottom.store(bottom + 1, relaxed);
      return false;
    }

    Continuation * peek() noexcept
    {
      u64 bottom = _bottom.load(relaxed) - 1;
      ring_buffer * buffer = _buffer.load(relaxed);
      return buffer->get(bottom);
    }

    Continuation * steal() noexcept
    {
      u64 top = _top.load(acquire);
      std::atomic_thread_fence(seq_cst);
      u64 bottom = _bottom.load(acquire);
      ring_buffer * buffer = _buffer.load(consume);
      if (top < bottom)
      {
        Continuation * x = buffer->get(top);
        if (!_top.compare_exchange_strong(top, top + 1, seq_cst, relaxed))
        {
          return nullptr;
        }
        return x;
      }
      return nullptr;
    }

    Continuation * pop_top() noexcept
    {
      u64 top = _top.load(relaxed);
      u64 bottom = _bottom.load(relaxed);
      auto buffer = _buffer.load(relaxed);
      Continuation * x = nullptr;
      if (top < bottom)
      {
        x = buffer->get(top);
        _top.store(top + 1, relaxed);
      }
      if (auto garbage = reclaim())
      {
        delete garbage;
      }
      return x;
    }

    ring_buffer * reclaim() noexcept
    {
      u64 bottom = _bottom.load(relaxed);
      u64 top = _top.load(relaxed);
      auto buffer = _buffer.load(relaxed);
      u64 capacity = buffer->capacity();
      if (bottom < top + (capacity >> 2))
      {
        if (auto next = buffer->shrink(top, bottom))
        {
          _buffer.store(next, relaxed);
          return buffer;
        }
      }
      return nullptr;
    }
  };

}


#endif // TYPON_FUNDAMENTAL_STACK_HPP_INCLUDED
